import datetime
import json
import logging
import os
import pathlib
import typing

import click

from np_aind_metadata import common, utils
from np_aind_metadata.np import dynamic_routing_task as dynamic_routing_task_np
from np_aind_metadata.update import dynamic_routing_task as dynamic_routing_task_update

logger = logging.getLogger(__name__)


NPEXP_PATH_ROOT = os.getenv("NPEXP_PATH_ROOT")
now = datetime.datetime.now()


@click.group()
@click.option("--debug/--no-debug", default=False)
@click.option("--log-file", default=False, is_flag=True)
def cli(debug: bool, log_file: bool) -> None:
    click.echo(f"Debug mode is {'on' if debug else 'off'}")
    if debug:
        dynamic_routing_task_np.logger.setLevel(logging.DEBUG)
        dynamic_routing_task_update.logger.setLevel(logging.DEBUG)
        utils.logger.setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)

    if log_file:
        handler = logging.FileHandler(
            now.strftime("np-aind-metadata_%Y-%m-%d-%H-%M-%S") + ".log"
        )
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        logger.addHandler(handler)
        dynamic_routing_task_np.logger.addHandler(handler)
        dynamic_routing_task_update.logger.addHandler(handler)
        utils.logger.addHandler(handler)


@cli.command()
@click.argument(
    "rig-name",
    type=str,
)
@click.argument(
    "modification-date-str",
    type=str,
)
@click.option(
    "--output-directory",
    type=pathlib.Path,
    default=pathlib.Path("./"),
)
def init_rig(
    rig_name: str,
    modification_date_str: str,
    output_directory: pathlib.Path,
) -> None:
    if rig_name not in ["NP0", "NP1", "NP2", "NP3"]:
        raise ValueError(f"Invalid rig name: {rig_name}")

    modification_date = datetime.datetime.strptime(
        modification_date_str, "%Y-%m-%d"
    ).date()

    output_path = dynamic_routing_task_np.init_rig_from_np_config(
        rig_name,
        modification_date,
        output_directory,
    )
    click.echo(f"Rig json generated at: {output_path}")


@cli.command()
@click.argument(
    "rig-path",
    type=pathlib.Path,
)
@click.argument(
    "session-dir",
    type=pathlib.Path,
)
@click.option(
    "--output-dir",
    type=pathlib.Path,
    default=pathlib.Path("./"),
)
@click.option(
    "--include",
    multiple=True,
)
def update_dynamic_routing_rig(
    rig_path: pathlib.Path,
    session_dir: pathlib.Path,
    output_dir: pathlib.Path,
    include: typing.Optional[list[str]],
) -> None:
    rig_path = utils.fix_allen_path(rig_path)
    session_dir = utils.fix_allen_path(session_dir)
    logger.debug("Rig path: %s" % rig_path)
    logger.debug("Session dir: %s" % session_dir)
    logger.debug("Include: %s" % json.dumps(include))
    output_path = dynamic_routing_task_np.update_rig_from_session_dir(
        rig_path,
        session_dir,
        output_dir=output_dir,
        include=include,
    )
    click.echo(f"Updated rig json generated at: {output_path}")


@cli.command()
@click.argument("rig-path", type=pathlib.Path)
@click.argument("session-root-dirs", type=pathlib.Path, nargs=-1)
@click.option("--output-dir", type=pathlib.Path, default=pathlib.Path("./"))
@click.option(
    "--include",
    multiple=True,
)
def batch_generate_np(
    rig_path: pathlib.Path,
    session_root_dirs: list[pathlib.Path],
    output_dir: pathlib.Path,
    include: typing.Optional[list[str]],
) -> None:
    rig_path = utils.fix_allen_path(rig_path)
    session_root_dirs = [utils.fix_allen_path(p) for p in session_root_dirs]
    logger.debug(f"Rig path: {rig_path}")
    logger.debug(f"Session root dirs: {session_root_dirs}")
    for session_dir in dynamic_routing_task_np.get_sorted_session_dirs(
        *session_root_dirs
    ):
        output_path = dynamic_routing_task_np.update_rig_from_session_dir(
            rig_path,
            session_dir,
            output_dir=output_dir,
            include=include,
        )
        logger.debug(f"Updated rig json generated at: {output_path}")


@cli.command()
@click.argument("rig-path", type=pathlib.Path)
@click.argument("sorted-sessions-path", type=pathlib.Path)
@click.option("--output-dir", type=pathlib.Path, default=pathlib.Path("./"))
def backfill_generate(
    rig_path: pathlib.Path,
    sorted_sessions_path: pathlib.Path,
    output_dir: pathlib.Path,
) -> None:
    rig_path = utils.fix_allen_path(rig_path)
    session_dir_infos = [
        (pathlib.Path(path), subject_id, is_surface)
        for _, path, subject_id, is_surface in json.loads(
            sorted_sessions_path.read_text()
        )
    ]
    logger.debug(f"Rig path: {rig_path}")
    for session_dir, subject_id, is_surface in session_dir_infos:
        update_context = dynamic_routing_task_np.update_rig_from_session_dir(
            rig_path,
            session_dir,
            output_dir=output_dir,
            include=["open_ephys_context"],
            id_prefix=subject_id,
            id_suffix="surface_channels" if is_surface else None,
        )
        logger.debug(f"Updated rig json generated at: {update_context.updated}")
        if not update_context.updated:
            logger.error("Failed to update rig json.")
            continue
        rig_path = update_context.updated


@cli.command()
@click.argument("rig-path", type=pathlib.Path)
@click.option("--mvr-source", type=pathlib.Path)
@click.option("--sync-source", type=pathlib.Path)
@click.option("--output-path", type=pathlib.Path)
def manual_update(
    rig_path: pathlib.Path,
    mvr_source: pathlib.Path | None,
    sync_source: pathlib.Path | None,
    output_path: pathlib.Path | None,
) -> None:
    context = common.DynamicRoutingTaskUpdateContext(
        source=rig_path,
    )
    if mvr_source:
        context.mvr_context = common.MVRContext(
            source=mvr_source,
        )
    if sync_source:
        context.sync_context = common.SyncContext(
            source=sync_source,
        )
    if output_path is None:
        output_path = rig_path
    updated_context = dynamic_routing_task_update.update_rig(context, output_path)
    click.echo(f"Updated rig json generated at: {updated_context.updated}")


if __name__ == "__main__":
    cli()



"""ETL for the dynamic routing task."""

import logging
from datetime import date, datetime
from pathlib import Path
from typing import Optional

import numpy as np
from aind_data_schema.core.rig import Rig
from aind_data_schema.models.devices import (
    Calibration,
    DAQChannel,
    DaqChannelType,
    Disc,
    Monitor,
)
from aind_data_schema.models.units import FrequencyUnit, SizeUnit
from aind_metadata_mapper.neuropixels.neuropixels_rig import (
    NeuropixelsRigContext,
    NeuropixelsRigEtl,
)

from np_aind_metadata import utils

logger = logging.getLogger(__name__)


class ExtractContext(NeuropixelsRigContext):
    """Extract context for DynamicRoutingTaskRigEtl."""

    version: Optional[str]
    reward_line: Optional[tuple[int, int]]
    reward_sound_line: Optional[tuple[int, int]]
    lick_line: Optional[tuple[int, int]]
    frame_signal_line: Optional[tuple[int, int]]
    acquisition_signal_line: Optional[tuple[int, int]]
    opto_channels: Optional[dict[str, list[int]]]
    galvo_channels: Optional[tuple[int, int]]
    monitor_distance: Optional[float]
    monitor_size: Optional[tuple[int, int]]
    wheel_radius: Optional[float]
    sound_calibration_fit: Optional[tuple[float, float, float]]
    solenoid_open_time: Optional[float]


# dynamic routing task has two slashes before commit hash
SUPPORTED_VERSIONS = [
    (
        b"https://raw.githubusercontent.com/samgale/DynamicRoutingTask"
        b"//9ea009a6c787c0049648ab9a93eb8d9df46d3f7b/DynamicRouting1.py"
    ),
]


class DynamicRoutingTaskRigEtl(NeuropixelsRigEtl):
    """DynamicRouting rig ETL class. Extracts information from task output
    file.
    """

    def __init__(
        self,
        input_source: Path,
        output_directory: Path,
        task_source: Path,
        monitor_name: str = "Stim",
        speaker_name: str = "Speaker",
        behavior_daq_name: str = "Behavior",
        behavior_sync_daq_name: str = "BehaviorSync",
        opto_daq_name: str = "Opto",
        reward_delivery_name: str = "Reward delivery",
        sound_calibration_date: Optional[date] = None,
        reward_calibration_date: Optional[date] = None,
        **kwargs,
    ) -> None:
        """Class constructor for Dynamic Routing rig etl class."""
        super().__init__(input_source, output_directory, **kwargs)
        self.task_source = task_source
        self.monitor_name = monitor_name
        self.speaker_name = speaker_name
        self.behavior_daq_name = behavior_daq_name
        self.behavior_sync_daq_name = behavior_sync_daq_name
        self.opto_daq_name = opto_daq_name
        self.reward_delivery_name = reward_delivery_name
        self.sound_calibration_date = sound_calibration_date
        self.reward_calibration_date = reward_calibration_date

    def _extract(self) -> ExtractContext:
        """Extracts DynamicRouting-related task information from task
        output."""
        task = utils.load_hdf5(self.task_source)
        return ExtractContext(
            current=super()._extract(),
            version=utils.extract_hdf5_value(task, ["githubTaskScript"]),
            reward_line=utils.extract_hdf5_value(task, ["rewardLine"]),
            reward_sound_line=utils.extract_hdf5_value(task, ["rewardSoundLine"]),
            lick_line=utils.extract_hdf5_value(task, ["lickLine"]),
            frame_signal_line=utils.extract_hdf5_value(task, ["frameSignalLine"]),
            acquisition_signal_line=utils.extract_hdf5_value(
                task, ["acquisitionSignalLine"]
            ),
            opto_channels=utils.extract_hdf5_value(task, ["optoChannels"]),
            galvo_channels=utils.extract_hdf5_value(task, ["galvoChannels"]),
            monitor_distance=utils.extract_hdf5_value(task, ["monDistance"]),
            monitor_size=utils.extract_hdf5_value(task, ["monSizePix"]),
            wheel_radius=utils.extract_hdf5_value(task, ["wheelRadius"]),
            sound_calibration_fit=utils.extract_hdf5_value(
                task, ["soundCalibrationFit"]
            ),
            solenoid_open_time=utils.extract_hdf5_value(task, ["solenoidOpenTime"]),
        )

    def _transform_daq_channels(
        self,
        extracted_source: ExtractContext,
        daq_name: str,
        channels: list[DAQChannel],
    ) -> None:
        """Updates channel settings for a given daq."""
        for daq in extracted_source.current.daqs:
            if daq.name == daq_name:
                logger.debug("Updating daq=%s." % daq_name)
                for channel in channels:
                    utils.find_replace_or_append(
                        daq.channels,
                        [("channel_name", channel.channel_name)],
                        channel,
                    )

    def _transform_behavior_daq(self, extracted_source: ExtractContext) -> None:
        """Updates rig model with DynamicRouting-related behavior daq
        information."""
        behavior_daq_channels = []
        if extracted_source.reward_line is not None:
            logger.debug(
                "Extracted reward line port, channel: %s" % self.behavior_daq_name
            )
            behavior_daq_channels.append(
                DAQChannel(
                    device_name=self.behavior_daq_name,
                    channel_name="solenoid",
                    channel_type=DaqChannelType.DO,
                    port=extracted_source.reward_line[0],
                    channel_index=extracted_source.reward_line[1],
                )
            )

        if extracted_source.reward_sound_line is not None:
            logger.debug(
                "Extracted reward sound line port, channel: %s, %s"
                % extracted_source.reward_sound_line
            )
            behavior_daq_channels.append(
                DAQChannel(
                    device_name=self.behavior_daq_name,
                    channel_name="reward_sound",
                    channel_type=DaqChannelType.DO,
                    port=extracted_source.reward_sound_line[0],
                    channel_index=extracted_source.reward_sound_line[1],
                )
            )

        if extracted_source.lick_line is not None:
            logger.debug(
                "Extracted lick line on port, channel: %s, %s"
                % extracted_source.lick_line
            )
            behavior_daq_channels.append(
                DAQChannel(
                    device_name=self.behavior_daq_name,
                    channel_name="lick",
                    channel_type=DaqChannelType.DI,
                    port=extracted_source.lick_line[0],
                    channel_index=extracted_source.lick_line[1],
                )
            )

        if behavior_daq_channels:
            self._transform_daq_channels(
                extracted_source, self.behavior_daq_name, behavior_daq_channels
            )

    def _transform_behavior_sync_daq(self, extracted_source: ExtractContext) -> None:
        """Updates rig model with DynamicRouting-related behavior sync daq
        information."""
        behavior_sync_daq_channels = []
        if extracted_source.frame_signal_line is not None:
            logger.debug(
                "Extracted frame signal port, channel: %s, %s"
                % extracted_source.frame_signal_line
            )
            behavior_sync_daq_channels.append(
                DAQChannel(
                    device_name=self.behavior_sync_daq_name,
                    channel_name="stim_frame",
                    channel_type=DaqChannelType.DO,
                    port=extracted_source.frame_signal_line[0],
                    channel_index=extracted_source.frame_signal_line[1],
                )
            )

        if extracted_source.acquisition_signal_line is not None:
            logger.debug(
                "Extracted aquisition port, channel: %s, %s"
                % extracted_source.acquisition_signal_line
            )
            behavior_sync_daq_channels.append(
                DAQChannel(
                    device_name=self.behavior_sync_daq_name,
                    channel_name="stim_running",
                    channel_type=DaqChannelType.DO,
                    port=extracted_source.acquisition_signal_line[0],
                    channel_index=extracted_source.acquisition_signal_line[1],
                )
            )

        if behavior_sync_daq_channels:
            self._transform_daq_channels(
                extracted_source,
                self.behavior_sync_daq_name,
                behavior_sync_daq_channels,
            )

    def _transform_opto_daq(self, extracted_source: ExtractContext) -> None:
        """Updates rig model with DynamicRouting-related opto daq
        information."""
        opto_daq_channels = []
        if extracted_source.opto_channels is not None:
            logger.debug("Extracted opto channels: %s" % extracted_source.opto_channels)
            if extracted_source.opto_channels:
                for device_name in sorted(extracted_source.opto_channels):
                    for idx, channel in enumerate(
                        extracted_source.opto_channels[device_name]
                    ):
                        opto_daq_channels.append(
                            DAQChannel(
                                device_name=self.opto_daq_name,
                                channel_name=(
                                    f"{self.opto_daq_name} " f"{device_name} #{idx}"
                                ),
                                channel_type=DaqChannelType.AO,
                                channel_index=channel,
                            )
                        )

        if extracted_source.galvo_channels is not None:
            logger.debug(
                "Extracted galvo channels x,y: %s, %s" % extracted_source.galvo_channels
            )
            opto_daq_channels.extend(
                [
                    DAQChannel(
                        device_name=self.opto_daq_name,
                        channel_name=f"{self.opto_daq_name} galvo x",
                        channel_type=DaqChannelType.AO,
                        channel_index=extracted_source.galvo_channels[0],
                        sample_rate=2000,
                        sample_rate_unit=FrequencyUnit.HZ,
                    ),
                    DAQChannel(
                        device_name=self.opto_daq_name,
                        channel_name=f"{self.opto_daq_name} galvo y",
                        channel_type=DaqChannelType.AO,
                        channel_index=extracted_source.galvo_channels[1],
                        sample_rate=2000,
                        sample_rate_unit=FrequencyUnit.HZ,
                    ),
                ]
            )

        if opto_daq_channels:
            self._transform_daq_channels(
                extracted_source, self.opto_daq_name, opto_daq_channels
            )

    def _transform_calibrations(self, extracted_source: ExtractContext) -> None:
        """Updates rig model with DynamicRouting-related calibration
        information."""
        default_calibration_date = datetime.now()

        # sound
        if extracted_source.sound_calibration_fit is not None:
            logger.debug("Updating sound calibration")
            utils.find_replace_or_append(
                extracted_source.current.calibrations,
                [
                    ("device_name", self.speaker_name),
                ],
                Calibration(
                    calibration_date=self.sound_calibration_date
                    or default_calibration_date,
                    device_name=self.speaker_name,
                    input={
                        "a": extracted_source.sound_calibration_fit[0],
                        "b": extracted_source.sound_calibration_fit[1],
                        "c": extracted_source.sound_calibration_fit[2],
                    },
                    output={},
                    description=(
                        "sound_volume = log(1 - ((dB - c) / a)) / b;"
                        "dB is sound pressure"
                    ),
                    notes=("Calibration date is a placeholder. "),
                ),
            )

        # water
        if extracted_source.solenoid_open_time is not None:
            logger.debug("Updating reward delivery calibration")
            utils.find_replace_or_append(
                extracted_source.current.calibrations,
                [
                    ("device_name", self.reward_delivery_name),
                ],
                Calibration(
                    calibration_date=self.reward_calibration_date
                    or default_calibration_date,
                    device_name=self.reward_delivery_name,
                    input={},
                    output={
                        "solenoid_open_time": (extracted_source.solenoid_open_time),
                    },
                    description=(
                        "solenoid open time (ms) = slope * expected water "
                        "volume (mL) + intercept"
                    ),
                    notes=("Calibration date is a placeholder."),
                ),
            )

    def _transform(self, extracted_source: ExtractContext) -> Rig:
        """Updates rig model with DynamicRouting-related task information."""
        if extracted_source.version is not None:
            if extracted_source.version not in SUPPORTED_VERSIONS:
                logger.warning(
                    f"Unsupported task version: {extracted_source.version}",
                )

        # monitor information
        if (
            extracted_source.monitor_distance is not None
            or extracted_source.monitor_size is not None
        ):
            for idx, device in enumerate(extracted_source.current.stimulus_devices):
                if device.name == self.monitor_name and device.device_type == "Monitor":
                    if extracted_source.monitor_distance is not None:
                        device.viewing_distance = float(
                            extracted_source.monitor_distance
                        )
                        device.viewing_distance_unit = SizeUnit.CM

                    if extracted_source.monitor_size is not None:
                        width, height = extracted_source.monitor_size
                        if not np.isnan(width) and not np.isnan(height):
                            device.width = int(width)
                            device.height = int(height)
                            device.size_unit = SizeUnit.PX

                    extracted_source.current.stimulus_devices[idx] = (
                        Monitor.model_validate(device.__dict__)
                    )
                    break

        # wheel info
        if extracted_source.wheel_radius is not None:
            logger.debug("Updating wheel information")
            extracted_source.current.mouse_platform.radius = (
                extracted_source.wheel_radius
            )
            extracted_source.current.mouse_platform.radius_unit = SizeUnit.CM
            extracted_source.current.mouse_platform = Disc.model_validate(
                extracted_source.current.mouse_platform.__dict__
            )
        # daqs
        self._transform_behavior_daq(extracted_source)
        self._transform_behavior_sync_daq(extracted_source)
        self._transform_opto_daq(extracted_source)
        # calibrations
        self._transform_calibrations(extracted_source)

        return super()._transform(extracted_source.current)



import pathlib

from aind_data_schema.core import rig
from aind_data_schema.models import coordinates, devices, organizations

from np_aind_metadata import common

COPA_NOTES = (
    "The rotation matrix is represented as: a,b,c,d,e,f,g,h,i. Wherein a, b, "
    "c correspond to the first row of the matrix. The translation matrix is "
    "represented as: x,y,z."
)


def init(
    config: common.DynamicRoutingRigDetails,
    output_directory: pathlib.Path = pathlib.Path("."),
) -> pathlib.Path:
    """Initializes a rig model for the dynamic routing project.

    Notes
    -----
    - rig_id is expected to be in the format:
        <ROOM NAME>_<RIG NAME>_<MODIFICATION DATE>
    - The DR task does not set the brightness and contrast of the monitor.
     These are hardcoded and assumed to be static.
    """
    rig_id = f"{config.rig_name}_{config.modification_date.strftime('%y%m%d')}"
    if config.room_name is not None:
        rig_id = f"{config.room_name}_{rig_id}"

    shared_camera_props = {
        "detector_type": "Camera",
        "model": "G-032",
        "max_frame_rate": 102,
        "sensor_width": 7400,
        "sensor_height": 7400,
        "size_unit": devices.SizeUnit.NM,
        "notes": "Max frame rate is at maximum resolution.",
        "cooling": devices.Cooling.NONE,
    }

    shared_camera_assembly_relative_position_props = {
        "device_origin": (
            "Located on the face of the lens mounting surface at its center, "
            "ie. just ahead of the camera sensor."
        ),
        "device_axes": [
            coordinates.Axis(
                name=coordinates.AxisName.X,
                direction=(
                    "Positive is from the center of the sensor towards the "
                    "left of the assembly, from the subject's perspective."
                ),
            ),
            coordinates.Axis(
                name=coordinates.AxisName.Y,
                direction=(
                    "Positive is from the center of the sensor towards the "
                    "bottom of the assembly."
                ),
            ),
            coordinates.Axis(
                name=coordinates.AxisName.Z,
                direction=("Positive is from the sensor towards the subject."),
            ),
        ],
        "notes": COPA_NOTES,
    }

    ephys_assemblies = []
    for assembly_letter in ["A", "B", "C", "D", "E", "F"]:
        serial_number = None
        for manipulator_info in config.manipulator_infos:
            if manipulator_info.assembly_name == f"Ephys Assembly {assembly_letter}":
                serial_number = manipulator_info.serial_number
                break
        ephys_assemblies.append(
            rig.EphysAssembly(
                name=f"Ephys Assembly {assembly_letter}",
                manipulator=devices.Manipulator(
                    name=f"Ephys Assembly {assembly_letter} Manipulator",
                    manufacturer=organizations.Organization.NEW_SCALE_TECHNOLOGIES,
                    model="06591-M-0004",
                    serial_number=serial_number,
                ),
                probes=[
                    devices.EphysProbe(
                        name=f"Probe{assembly_letter}",
                        probe_model="Neuropixels 1.0",
                        manufacturer=organizations.Organization.IMEC,
                    )
                ],
            )
        )

    model = rig.Rig(
        rig_id=rig_id,
        modification_date=config.modification_date,
        modalities=[
            rig.Modality.BEHAVIOR_VIDEOS,
            rig.Modality.BEHAVIOR,
            rig.Modality.ECEPHYS,
        ],
        mouse_platform=devices.Disc(
            name="Mouse Platform",
            radius="4.69",
            radius_unit="centimeter",
            notes=(
                "Radius is the distance from the center of the wheel to the " "mouse."
            ),
        ),
        stimulus_devices=[
            devices.Monitor(
                name="Stim",
                model="PA248",
                manufacturer=organizations.Organization.ASUS,
                width=1920,
                height=1200,
                size_unit="pixel",
                viewing_distance=15.3,
                viewing_distance_unit="centimeter",
                refresh_rate=60,
                brightness=43,
                contrast=50,
                position=coordinates.RelativePosition(
                    device_position_transformations=[
                        coordinates.Rotation3dTransform(
                            rotation=[
                                -0.80914,
                                -0.58761,
                                0,
                                -0.12391,
                                0.17063,
                                0.97751,
                                0.08751,
                                -0.12079,
                                0.02298,
                            ],
                        ),
                        coordinates.Translation3dTransform(
                            translation=[0.08751, -0.12079, 0.02298]
                        ),
                    ],
                    device_origin=(
                        "Located at the center of the screen. Right and left "
                        "conventions are relative to the screen side of the "
                        "monitor, ie. from the subject's perspective."
                    ),
                    device_axes=[
                        coordinates.Axis(
                            name=coordinates.AxisName.X,
                            direction=(
                                "Positive is from the center towards the "
                                "right of the screen, from the subject's "
                                "perspective."
                            ),
                        ),
                        coordinates.Axis(
                            name=coordinates.AxisName.Y,
                            direction=(
                                "Positive is from the center towards the top "
                                "of the screen."
                            ),
                        ),
                        coordinates.Axis(
                            name=coordinates.AxisName.Z,
                            direction=(
                                "Positive is from the front of the screen "
                                "towards the subject."
                            ),
                        ),
                    ],
                    notes=COPA_NOTES,
                ),
            ),
            devices.Speaker(
                name="Speaker",
                manufacturer=organizations.Organization.ISL,
                model="SPK-I-81345",
                position=coordinates.RelativePosition(
                    device_position_transformations=[
                        coordinates.Rotation3dTransform(
                            rotation=[
                                -0.82783,
                                -0.4837,
                                -0.28412,
                                -0.55894,
                                0.75426,
                                0.34449,
                                0.04767,
                                0.44399,
                                -0.89476,
                            ],
                        ),
                        coordinates.Translation3dTransform(
                            translation=[-0.00838, -0.09787, 0.18228]
                        ),
                    ],
                    device_origin=(
                        "Located on the front mounting flange face. Right "
                        "and left conventions are relative to the front side "
                        "of the speaker, ie. from the subject's perspective."
                    ),
                    device_axes=[
                        coordinates.Axis(
                            name=coordinates.AxisName.X,
                            direction=(
                                "Positive is from the center of the speaker "
                                "towards the right of the speaker, from the "
                                "subject's perspective."
                            ),
                        ),
                        coordinates.Axis(
                            name=coordinates.AxisName.Y,
                            direction=(
                                "Positive is from the center towards the top "
                                "of the speaker."
                            ),
                        ),
                        coordinates.Axis(
                            name=coordinates.AxisName.Z,
                            direction=(
                                "Positive is from the speaker towards the " "subject."
                            ),
                        ),
                    ],
                    notes=COPA_NOTES
                    + (
                        " Speaker to be mounted with the X axis pointing to "
                        "the right when viewing the speaker along the Z axis"
                    ),
                ),
            ),
            devices.RewardDelivery(
                reward_spouts=[
                    devices.RewardSpout(
                        name="Reward Spout",
                        manufacturer=organizations.Organization.HAMILTON,
                        model="8649-01 Custom",
                        spout_diameter=0.672,
                        spout_diameter_unit="millimeter",
                        side=devices.SpoutSide.CENTER,
                        solenoid_valve=devices.Device(
                            name="Solenoid Valve",
                            device_type="Solenoid Valve",
                            manufacturer=organizations.Organization.NRESEARCH,
                            model="161K011",
                            notes="Model number is product number.",
                        ),
                        lick_sensor=devices.Device(
                            name="Lick Sensor",
                            device_type="Lick Sensor",
                            manufacturer=organizations.Organization.OTHER,
                        ),
                        lick_sensor_type=devices.LickSensorType.PIEZOELECTIC,
                        notes=(
                            "Spout diameter is for inner diameter. "
                            "Outer diameter is 1.575mm. "
                        ),
                    ),
                ]
            ),
        ],
        ephys_assemblies=ephys_assemblies,
        light_sources=[
            devices.LightEmittingDiode(
                manufacturer=organizations.Organization.OTHER,
                name="Face forward LED",
                model="LZ4-40R308-0000",
                wavelength=740,
                wavelength_unit=devices.SizeUnit.NM,
            ),
            devices.LightEmittingDiode(
                manufacturer=organizations.Organization.OTHER,
                name="Body LED",
                model="LZ4-40R308-0000",
                wavelength=740,
                wavelength_unit=devices.SizeUnit.NM,
            ),
            devices.LightEmittingDiode(
                manufacturer=organizations.Organization.OSRAM,
                name="Eye LED",
                model="LZ4-40R608-0000",
                wavelength=850,
                wavelength_unit=devices.SizeUnit.NM,
            ),
            devices.Laser(
                name="Laser #0",
                manufacturer=organizations.Organization.VORTRAN,
                wavelength=488.0,
                model="Stradus 488-50",
                wavelength_unit="nanometer",
            ),
            devices.Laser(
                name="Laser #1",
                manufacturer=organizations.Organization.VORTRAN,
                wavelength=633.0,
                model="Stradus 633-80",
                wavelength_unit="nanometer",
            ),
        ],
        cameras=[
            devices.CameraAssembly(
                name="Front",
                camera_target="Face forward",
                camera=devices.Camera(
                    name="Front camera",
                    manufacturer=organizations.Organization.ALLIED,
                    chroma="Monochrome",
                    data_interface="Ethernet",
                    computer_name=config.mon_computer_name,
                    **shared_camera_props,
                ),
                filter=devices.Filter(
                    name="Front filter",
                    manufacturer=organizations.Organization.SEMROCK,
                    model="FF01-715_LP-25",
                    filter_type=devices.FilterType.LONGPASS,
                ),
                lens=devices.Lens(
                    name="Front lens",
                    manufacturer=organizations.Organization.EDMUND_OPTICS,
                    focal_length=8.5,
                    focal_length_unit="millimeter",
                    model="86604",
                ),
                position=coordinates.RelativePosition(
                    device_position_transformations=[
                        coordinates.Rotation3dTransform(
                            rotation=[
                                -0.17365,
                                0.98481,
                                0,
                                0.44709,
                                0.07883,
                                -0.89101,
                                -0.87747,
                                -0.15472,
                                -0.45399,
                            ]
                        ),
                        coordinates.Translation3dTransform(
                            translation=[0.154, 0.03078, 0.06346],
                        ),
                    ],
                    **shared_camera_assembly_relative_position_props,
                ),
            ),
            devices.CameraAssembly(
                name="Side",
                camera_target="Body",
                camera=devices.Camera(
                    name="Side camera",
                    manufacturer=organizations.Organization.ALLIED,
                    chroma="Monochrome",
                    data_interface="Ethernet",
                    computer_name=config.mon_computer_name,
                    **shared_camera_props,
                ),
                filter=devices.Filter(
                    name="Side filter",
                    manufacturer=organizations.Organization.SEMROCK,
                    model="FF01-747/33-25",
                    filter_type=devices.FilterType.BANDPASS,
                ),
                lens=devices.Lens(
                    name="Side lens",
                    manufacturer=organizations.Organization.NAVITAR,
                    focal_length=6.0,
                    focal_length_unit="millimeter",
                ),
                position=coordinates.RelativePosition(
                    device_position_transformations=[
                        coordinates.Rotation3dTransform(
                            rotation=[-1, 0, 0, 0, 0, -1, 0, -1, 0]
                        ),
                        coordinates.Translation3dTransform(
                            translation=[-0.03617, 0.23887, -0.02535],
                        ),
                    ],
                    **shared_camera_assembly_relative_position_props,
                ),
            ),
            devices.CameraAssembly(
                name="Eye",
                camera_target="Eye",
                camera=devices.Camera(
                    name="Eye camera",
                    manufacturer=organizations.Organization.ALLIED,
                    chroma="Monochrome",
                    data_interface="Ethernet",
                    computer_name=config.mon_computer_name,
                    **shared_camera_props,
                ),
                filter=devices.Filter(
                    name="Eye filter",
                    manufacturer=organizations.Organization.SEMROCK,
                    model="FF01-850/10-25",
                    filter_type=devices.FilterType.BANDPASS,
                ),
                lens=devices.Lens(
                    name="Eye lens",
                    manufacturer=organizations.Organization.INFINITY_PHOTO_OPTICAL,
                    focal_length=6.0,
                    focal_length_unit="millimeter",
                    model="213073",
                    notes="Model number is SKU.",
                ),
                position=coordinates.RelativePosition(
                    device_position_transformations=[
                        coordinates.Rotation3dTransform(
                            rotation=[
                                -0.5,
                                -0.86603,
                                0,
                                -0.366,
                                0.21131,
                                -0.90631,
                                0.78489,
                                -0.45315,
                                -0.42262,
                            ]
                        ),
                        coordinates.Translation3dTransform(
                            translation=[-0.14259, 0.06209, 0.09576],
                        ),
                    ],
                    **shared_camera_assembly_relative_position_props,
                ),
            ),
        ],
        daqs=[
            devices.DAQDevice(
                manufacturer=organizations.Organization.NATIONAL_INSTRUMENTS,
                name="Sync",
                computer_name=config.sync_computer_name,
                model="NI-6612",
                data_interface=devices.DataInterface.PCIE,
            ),
            devices.DAQDevice(
                manufacturer=organizations.Organization.NATIONAL_INSTRUMENTS,
                name="Behavior",
                computer_name=config.stim_computer_name,
                model="NI-6323",
                data_interface=devices.DataInterface.USB,
            ),
            devices.DAQDevice(
                manufacturer=organizations.Organization.NATIONAL_INSTRUMENTS,
                name="BehaviorSync",
                computer_name=config.stim_computer_name,
                model="NI-6001",
                data_interface=devices.DataInterface.PCIE,
            ),
            devices.DAQDevice(
                manufacturer=organizations.Organization.NATIONAL_INSTRUMENTS,
                name="Opto",
                computer_name=config.stim_computer_name,
                model="NI-9264",
                data_interface=devices.DataInterface.ETH,
            ),
        ],
        detectors=[
            devices.Detector(
                name="vsync photodiode",
                model="PDA25K",
                manufacturer=organizations.Organization.THORLABS,
                data_interface=devices.DataInterface.OTHER,
                notes="Data interface is unknown.",
                detector_type=devices.DetectorType.OTHER,
                cooling=devices.Cooling.AIR,
            ),
        ],
        calibrations=[],
        additional_devices=[
            devices.Detector(
                name="microphone",
                manufacturer=organizations.Organization.DODOTRONIC,
                model="MOM",
                data_interface=devices.DataInterface.OTHER,
                notes="Data interface is unknown.",
                detector_type=devices.DetectorType.OTHER,
                cooling=devices.Cooling.AIR,
            ),
            devices.AdditionalImagingDevice(
                name="Galvo x",
                imaging_device_type=devices.ImagingDeviceType.GALVO,
            ),
            devices.AdditionalImagingDevice(
                name="Galvo y",
                imaging_device_type=devices.ImagingDeviceType.GALVO,
            ),
        ],
        rig_axes=[
            coordinates.Axis(
                name=coordinates.AxisName.X,
                direction=(
                    "The world horizontal. Lays on the Mouse Sagittal Plane. "
                    "Positive direction is towards the nose of the mouse. "
                ),
            ),
            coordinates.Axis(
                name=coordinates.AxisName.Y,
                direction=(
                    "Perpendicular to Y. Positive direction is "
                    "away from the nose of the mouse. "
                ),
            ),
            coordinates.Axis(
                name=coordinates.AxisName.Z,
                direction="Positive pointing up.",
            ),
        ],
        origin=coordinates.Origin.BREGMA,
        patch_cords=[
            devices.Patch(
                name="Patch Cord #1",
                manufacturer=organizations.Organization.THORLABS,
                model="SM450 Custom Length, FC/PC Ends",
                core_diameter=125.0,
                numerical_aperture=0.10,
                notes=("Numerical aperture is approximately between 0.10 and " "0.14."),
            ),
        ],
    )

    validated = rig.Rig.model_validate(model)
    validated.write_standard_file(output_directory)
    return output_directory / "rig.json"


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



import datetime
import json
import logging
import pathlib
import shutil
import tempfile
import typing

from aind_data_schema.core import rig
from aind_metadata_mapper.neuropixels import (
    mvr_rig,
    neuropixels_rig,
    open_ephys_rig,
    sync_rig,
)

from np_aind_metadata import common
from np_aind_metadata.update.dynamic_routing_task_etl import DynamicRoutingTaskRigEtl

logger = logging.getLogger(__name__)


def _run_neuropixels_rig_etl(
    etl_class: neuropixels_rig.NeuropixelsRigEtl,
    input_source: rig.Rig,
    output_dir: pathlib.Path,
    *etl_args,
    **etl_kwargs,
) -> pathlib.Path:
    """Utility for running a neuropixels rig ETL and continuing on error."""
    try:
        etl = etl_class(input_source, output_dir, *etl_args, **etl_kwargs)
        etl.run_job()
        return output_dir / "rig.json"
    except Exception:
        logger.debug("Error calling: %s" % etl_class, exc_info=True)
        return input_source


# cannot type hint due to np import failing in github actions
def scrape_update_context(
    path: pathlib.Path,
    context: common.DynamicRoutingTaskUpdateContext,
    manipulators: typing.Optional[list[common.ManipulatorInfo]] = None,
) -> common.DynamicRoutingTaskUpdateContext:
    """Scrapes the given directory for relevant files and updates
    UpdateContext.
    """
    # dr task
    try:
        task_source = next(path.glob("**/Dynamic*.hdf5"))
        context.task_context = common.DynamicRoutingTaskContext(
            source=task_source,
        )
    except StopIteration:
        logger.debug("No task output found.")

    # sync
    try:
        sync_source = next(path.glob("**/sync.yml"))
        context.sync_context = common.SyncContext(
            source=sync_source,
        )
    except StopIteration:
        logger.debug("No sync config found.")

    # mvr
    try:
        mvr_source = next(path.glob("**/mvr.ini"))
        context.mvr_context = common.MVRContext(
            source=mvr_source,
        )
    except StopIteration:
        logger.debug("No mvr config found.")

    # open ephys
    settings_sources = list(path.glob("**/settings.xml"))
    if len(settings_sources) < 1:
        logger.debug("No open ephys settings found.")

    context.open_ephys_context = common.OpenEphysContext(
        source=settings_sources,
        manipulators=manipulators,
    )

    return context


def update_rig(
    context: common.DynamicRoutingTaskUpdateContext,
    output_path: pathlib.Path = pathlib.Path("rig.json"),
    modification_date: typing.Optional[datetime.date] = None,
    include: typing.Optional[list[str]] = None,
) -> common.DynamicRoutingTaskUpdateContext:
    """Generates a new rig json file with the metadata from the given sources.

    Notes
    -----
    - If rig_source is None, the rig will be initialized with the default
     values.
    - *_source, if present will update various values in the rig model.
    """
    # build model in a temporary directory
    build_dir = pathlib.Path(tempfile.mkdtemp())
    build_source = build_dir / "rig.json"
    shutil.copy(context.source, build_source)
    if include:
        logger.debug("Including only: %s" % json.dumps(include))
        context = common.DynamicRoutingTaskUpdateContext(
            source=context.source, **{k: getattr(context, k) for k in include}
        )
        logger.debug("Updated context: %s" % context.model_dump_json())

    rig_model = rig.Rig.model_validate_json(build_source.read_text())

    if context.task_context:
        logger.debug("Updating rig model with dynamic routing task context.")
        _run_neuropixels_rig_etl(
            DynamicRoutingTaskRigEtl,
            build_source,
            build_dir,
            task_source=context.task_context.source,
            sound_calibration_date=rig_model.modification_date,
            reward_calibration_date=rig_model.modification_date,
        )

    if context.open_ephys_context:
        logger.debug("Updating rig model with open ephys context.")
        if context.open_ephys_context.manipulators:
            manipulators = [
                (
                    m.assembly_name,
                    m.serial_number,
                )
                for m in context.open_ephys_context.manipulators
            ]
        else:
            manipulators = []
        _run_neuropixels_rig_etl(
            open_ephys_rig.OpenEphysRigEtl,
            build_source,
            build_dir,
            open_ephys_settings_sources=context.open_ephys_context.source,
            probe_manipulator_serial_numbers=manipulators,
            modification_date=modification_date,
        )

    if context.sync_context:
        logger.debug("Updating rig model with sync source file.")
        _run_neuropixels_rig_etl(
            sync_rig.SyncRigEtl,
            build_source,
            build_dir,
            config_source=context.sync_context.source,
        )

    if context.mvr_context:
        logger.debug("Updating rig model with mvr source file.")
        if context.mvr_context.mapping:
            camera_aliases = {
                m.mvr_name: m.camera_assembly_name for m in context.mvr_context.mapping
            }
        else:
            camera_aliases = None
        _run_neuropixels_rig_etl(
            mvr_rig.MvrRigEtl,
            build_source,
            build_dir,
            mvr_config_source=context.mvr_context.source,
            mvr_mapping=camera_aliases,
        )

    shutil.copy2(build_source, output_path)
    context.updated = output_path
    return context


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



import ast
import datetime
import logging
import pathlib
import shutil
import tempfile
import typing

from aind_data_schema.core import rig

from np_aind_metadata import common
from np_aind_metadata.init import dynamic_routing_task as dynamic_routing_task_init
from np_aind_metadata.update import dynamic_routing_task as dynamic_routing_task_update

logger = logging.getLogger(__name__)

try:
    import np_config
    import np_session
except Exception:
    logger.error("Failed to import neuropixels-related dependencies.", exc_info=True)


RigName = typing.Literal["NP0", "NP1", "NP2", "NP3"]


def get_rig_room(rig_name: RigName) -> typing.Union[str, None]:
    rig_to_room = {
        "NP0": "325",
        "NP1": "325",
        "NP2": "327",
        "NP3": "342",
    }
    try:
        return rig_to_room[rig_name]
    except KeyError:
        logger.debug("No room found for rig: %s" % rig_name)
        return None


# cannot type hint due to np import failing in github actions
def _get_rig_config(rig_name: RigName):
    return np_config.Rig(ast.literal_eval(rig_name[-1]))


def get_manipulator_infos(
    rig_name: RigName,
) -> list[common.ManipulatorInfo]:
    return [
        common.ManipulatorInfo(
            assembly_name=f"Ephys Assembly {key}",
            serial_number=value,
        )
        for key, value in _get_rig_config(rig_name)
        .config["services"]["NewScaleCoordinateRecorder"]["probe_to_serial_number"]
        .items()
    ]


def is_session_exp_dir(path: pathlib.Path) -> bool:
    try:
        np_session.Session(path.stem)
        return True
    except Exception:
        return False


# cannot type hint due to np import failing in github actions
def init_rig_from_np_config(
    rig_name: RigName,
    modification_date: datetime.date,
    output_directory: pathlib.Path,
) -> pathlib.Path:
    rig_config = _get_rig_config(rig_name)
    rig_details = common.DynamicRoutingRigDetails(
        rig_name=rig_name,
        mon_computer_name=rig_config.Mon,
        stim_computer_name=rig_config.Stim,
        sync_computer_name=rig_config.Sync,
        room_name=get_rig_room(rig_name),
        modification_date=modification_date,
        manipulator_infos=get_manipulator_infos(rig_name),
    )
    return dynamic_routing_task_init.init(rig_details, output_directory)


def _fix_modification_date(prev: pathlib.Path, new: pathlib.Path) -> None:
    """Fixes an unintentional bug introduced with neuropixels etls."""
    prev_rig = rig.Rig.model_validate_json(prev.read_text())
    new_rig = rig.Rig.model_validate_json(new.read_text())
    prev_probe_serial_numbers = [
        ephys_assembly.probes[0].serial_number
        for ephys_assembly in prev_rig.ephys_assemblies
    ]
    new_probe_serial_numbers = [
        ephys_assembly.probes[0].serial_number
        for ephys_assembly in new_rig.ephys_assemblies
    ]
    if prev_probe_serial_numbers != new_probe_serial_numbers:
        logger.debug("Probe serial numbers changed. Not reverting modification date.")
        logger.debug(f"Previous: {prev_probe_serial_numbers}")
        logger.debug(f"New: {new_probe_serial_numbers}")
        return

    logger.debug("Probe serial numbers match.")
    logger.debug("Reverting modification date.")
    with tempfile.TemporaryDirectory() as temp_dir:
        new_rig.rig_id = prev_rig.rig_id
        new_rig.modification_date = prev_rig.modification_date
        new_rig.write_standard_file(temp_dir)
        shutil.copy2(pathlib.Path(temp_dir) / "rig.json", new)


def update_rig_from_session_dir(
    rig_path: pathlib.Path,
    session_dir: pathlib.Path,
    output_dir: pathlib.Path = pathlib.Path("./"),
    include: typing.Optional[list[str]] = None,
    id_prefix: typing.Optional[str] = None,
    id_suffix: typing.Optional[str] = None,
) -> common.DynamicRoutingTaskUpdateContext:
    context = common.DynamicRoutingTaskUpdateContext(
        source=rig_path,
    )
    context = dynamic_routing_task_update.scrape_update_context(session_dir, context)
    logger.debug("Scraped context: %s" % context.model_dump_json())
    session = np_session.Session(session_dir.stem)
    _id = session.date.strftime("%Y-%m-%d")
    if id_prefix:
        _id = f"{id_prefix}_{_id}"
    if id_suffix:
        _id = f"{_id}_{id_suffix}"
    logger.debug(f"ID: {_id}")
    output_path = output_dir / f"{_id}_rig.json"
    updated = dynamic_routing_task_update.update_rig(
        context,
        output_path=output_path,
        modification_date=session.date,
        include=include,
    )
    _fix_modification_date(updated.source, output_path)
    return updated


# using typing.Any due to np_session import failing in github actions
def get_session_dirs(
    npexp_path_root: pathlib.Path,
) -> typing.Generator[pathlib.Path, None, None]:
    for path in np_config.utils.normalize_path(npexp_path_root).iterdir():
        if is_session_exp_dir(path):
            yield path


def get_sorted_session_dirs(
    *session_path_roots: pathlib.Path,
) -> list[pathlib.Path]:
    all_session_dirs = []
    for session_path_root in session_path_roots:
        all_session_dirs.extend(list(get_session_dirs(session_path_root)))

    def get_session_dir_sort_key(path: pathlib.Path) -> datetime.date:
        return np_session.Session(path.stem).date

    return sorted(all_session_dirs, key=get_session_dir_sort_key)



"""Models and controlled language for aind-data-schema devices names.
"""

import datetime
import pathlib
import typing

import pydantic

# COMMON Device names

# GALVO_X = "Galvo x"
# GALVO_Y = "Galvo y"
# LASER_0 = "Laser 0"
# LASER_1 = "Laser 1"
# LASER_ASSEMBLY_0 = "Laser Assembly 0"
# LASER_ASSEMBLY_1 = "Laser Assembly 1"
# CAMERA_0 = "Camera 0"
# CAMERA_1 = "Camera 1"
# CAMERA_2 = "Camera 2"


class NeuropixelsETLContext(pydantic.BaseModel):
    """Base Context for Neuropixels ETL."""

    source: pathlib.Path


class DynamicRoutingTaskContext(NeuropixelsETLContext):
    """Context for the DynamicRoutingTask ETL."""


class SyncContext(NeuropixelsETLContext):
    """Context for the Sync ETL."""


class MVRCammeraAlias(pydantic.BaseModel):
    """Alias for the camera name in the MVR file."""

    mvr_name: str
    camera_assembly_name: str


class MVRContext(NeuropixelsETLContext):
    """Context for the MVR ETL."""

    mapping: list[MVRCammeraAlias] = [
        MVRCammeraAlias(mvr_name="Camera 1", camera_assembly_name="Side"),
        MVRCammeraAlias(mvr_name="Camera 2", camera_assembly_name="Eye"),
        MVRCammeraAlias(mvr_name="Camera 3", camera_assembly_name="Front"),
    ]


class ManipulatorInfo(pydantic.BaseModel):

    assembly_name: str
    serial_number: str


class OpenEphysContext(pydantic.BaseModel):
    """Context for the OpenEphys ETL."""

    source: list[pathlib.Path] = []
    manipulators: typing.Optional[list[ManipulatorInfo]] = None
    modification_date: typing.Optional[datetime.date] = None


class DynamicRoutingRigDetails(pydantic.BaseModel):
    """Rig model initialization details."""

    rig_name: str
    mon_computer_name: str
    stim_computer_name: str
    sync_computer_name: str
    room_name: typing.Optional[str] = None
    modification_date: datetime.date
    manipulator_infos: list[ManipulatorInfo] = []


class DynamicRoutingTaskUpdateContext(pydantic.BaseModel):
    """Update Context for the DynamicRoutingTask model."""

    source: typing.Union[pathlib.Path]
    updated: typing.Optional[pathlib.Path] = None
    task_context: typing.Optional[DynamicRoutingTaskContext] = None
    open_ephys_context: typing.Optional[OpenEphysContext] = None
    sync_context: typing.Optional[SyncContext] = None
    mvr_context: typing.Optional[MVRContext] = None



import logging
from pathlib import Path
from typing import Any, Union

from h5py import Dataset, File

logger = logging.getLogger(__name__)


def load_hdf5(h5_path: Path) -> File:
    """Load hdf5 file from path."""
    return File(h5_path, "r")


def extract_hdf5_value(h5_file: File, path: list[str]) -> Union[Any, None]:
    """Extract value from hdf5 file using a path. Path is a list of property
    names that are used to traverse the hdf5 file. A path of length greater
    than 1 is expected to point to a nested property.
    """
    try:
        value = None
        for part in path:
            value = h5_file[part]
    except KeyError as e:
        logger.warning(f"Key not found: {e}")
        return None

    if isinstance(value, Dataset):
        return value[()]
    else:
        return value


def find_replace_or_append(
    iterable: list[Any],
    filters: list[tuple[str, Any]],
    update: Any,
) -> None:
    """Find an item in a list of items that matches the filters and replace it.
    If no item is found, append.
    """
    for idx, obj in enumerate(iterable):
        if all(
            getattr(obj, prop_name, None) == prop_value
            for prop_name, prop_value in filters
        ):
            iterable[idx] = update
            break
    else:
        iterable.append(update)


def fix_allen_path(path: Path) -> Path:
    """Fix Allen path for Windows."""
    if path.parts[0] == "\x07llen":
        logger.debug(f"Fixing Allen path: {path}")
        return Path(r"\\allen", *path.parts[1:])
    return path
